"""
Lifeguard core settings
"""
import re
import sys
from os import environ


class AttributeNotFoundInSettings(Exception):
    """Raised when the entry not found in settings"""


class SettingsManager(object):
    def __init__(self, settings):
        self.settings = settings

        for entry in self.settings:
            setattr(self, "__{}".format(entry.lower()), self.__get_value(entry))

    def __get_value(self, name):
        options = self.settings[name]
        return environ.get(name, options["default"])

    def __search_dynamic_attribute(self, name):
        for entry in self.settings:
            if re.match(entry, name):
                return environ.get(name, self.settings[entry]["default"])

        raise AttributeNotFoundInSettings("{} not found".format(name))

    def read_value(self, name):
        """
        Read value from settings
        """
        if hasattr(self, "__{}".format(name.lower())):
            return getattr(self, "__{}".format(name.lower()))

        return self.__search_dynamic_attribute(name)


SETTINGS_MANAGER = SettingsManager(
    {
        "LIFEGUARD_SERVER_PORT": {
            "default": "5567",
            "description": "Lifeguard server port number",
        },
        "LIFEGUARD_DIRECTORY": {
            "default": ".",
            "description": "Location of validations and others resources",
        },
        "LIFEGUARD_LOG_LEVEL": {
            "default": "INFO",
            "description": "Sets the Lifeguard's core log level",
        },
        "LIFEGUARD_HTTP_PROXY": {
            "default": None,
            "description": "Proxy used to http calls",
        },
        "LIFEGUARD_HTTPS_PROXY": {
            "default": None,
            "description": "Proxy used to https calls",
        },
        "LIFEGUARD_PUBLIC_ADDRESS": {
            "default": "http://lolcahost:5567",
            "description": "Address to access dashboard",
        },
        "LIFEGUARD_EMAIL_SMTP_USER": {
            "default": "smtp_user",
            "description": "SMTP user used in email action",
        },
        "LIFEGUARD_EMAIL_SMTP_PASSWD": {
            "default": "smtp_passwd",
            "description": "SMTP passwd used in email action",
        },
        "LIFEGUARD_EMAIL_SMTP_SERVER": {
            "default": "smtp_server",
            "description": "SMTP server used in email action",
        },
        "LIFEGUARD_EMAIL_SMTP_PORT": {
            "default": "smtp_port",
            "description": "SMTP port used in email action",
        },
    }
)

LIFEGUARD_SERVER_PORT = SETTINGS_MANAGER.read_value("LIFEGUARD_SERVER_PORT")
LIFEGUARD_DIRECTORY = SETTINGS_MANAGER.read_value("LIFEGUARD_DIRECTORY")
LIFEGUARD_PUBLIC_ADDRESS = SETTINGS_MANAGER.read_value("LIFEGUARD_PUBLIC_ADDRESS")
LIFEGUARD_EMAIL_SMTP_USER = SETTINGS_MANAGER.read_value("LIFEGUARD_EMAIL_SMTP_USER")
LIFEGUARD_EMAIL_SMTP_PASSWD = SETTINGS_MANAGER.read_value("LIFEGUARD_EMAIL_SMTP_PASSWD")
LIFEGUARD_EMAIL_SMTP_SERVER = SETTINGS_MANAGER.read_value("LIFEGUARD_EMAIL_SMTP_SERVER")
LIFEGUARD_EMAIL_SMTP_PORT = SETTINGS_MANAGER.read_value("LIFEGUARD_EMAIL_SMTP_PORT")

LOG_LEVEL = SETTINGS_MANAGER.read_value("LIFEGUARD_LOG_LEVEL")
HTTP_PROXY = SETTINGS_MANAGER.read_value("LIFEGUARD_HTTP_PROXY")
HTTPS_PROXY = SETTINGS_MANAGER.read_value("LIFEGUARD_HTTPS_PROXY")



import unittest

from unittest.mock import patch, call

from lifeguard.settings import (
    LIFEGUARD_SERVER_PORT,
    LIFEGUARD_DIRECTORY,
    LOG_LEVEL,
    SETTINGS_MANAGER,
    SettingsManager,
    AttributeNotFoundInSettings,
)


class SettingsTest(unittest.TestCase):
    def test_server_port(self):
        self.assertEqual(LIFEGUARD_SERVER_PORT, "5567")
        self.assertEqual(
            SETTINGS_MANAGER.settings["LIFEGUARD_SERVER_PORT"]["description"],
            "Lifeguard server port number",
        )

    def test_log_level(self):
        self.assertEqual(LOG_LEVEL, "INFO")
        self.assertEqual(
            SETTINGS_MANAGER.settings["LIFEGUARD_LOG_LEVEL"]["description"],
            "Sets the Lifeguard's core log level",
        )

    def test_lifeguard_directory(self):
        self.assertEqual(LIFEGUARD_DIRECTORY, ".")
        self.assertEqual(
            SETTINGS_MANAGER.settings["LIFEGUARD_DIRECTORY"]["description"],
            "Location of validations and others resources",
        )

    def test_get_default_value_for_dynamic_attribute(self):
        settings = SettingsManager(
            {
                r"DYNAMIC_\w+_ATTRIBUTE": {
                    "default": "default_value",
                    "description": "dynammic descritpion",
                }
            }
        )

        self.assertEqual(settings.read_value("DYNAMIC_TEST_ATTRIBUTE"), "default_value")

    @patch("lifeguard.settings.environ")
    def test_get_value_for_dynamic_attribute_from_env(self, mock_environ):
        mock_environ.get.return_value = "from env"
        settings = SettingsManager(
            {
                r"DYNAMIC_\w+_ATTRIBUTE": {
                    "default": "default_value",
                    "description": "dynammic descritpion",
                }
            }
        )

        self.assertEqual(settings.read_value("DYNAMIC_TEST_ATTRIBUTE"), "from env")
        mock_environ.get.assert_has_calls(
            [
                call("DYNAMIC_\\w+_ATTRIBUTE", "default_value"),
                call("DYNAMIC_TEST_ATTRIBUTE", "default_value"),
            ]
        )

    def test_raise_exception_if_pattern_not_found(self):
        settings = SettingsManager(
            {
                r"DYNAMIC_\w+_ATTRIBUTE": {
                    "default": "default_value",
                    "description": "dynammic descritpion",
                }
            }
        )
        with self.assertRaises(AttributeNotFoundInSettings):
            settings.read_value("DYNAMIC__ATTRIBUTE")



"""
Lifeguard validation core
"""
import importlib
import os
import sys

from lifeguard.logger import lifeguard_logger as logger
from lifeguard.repositories import declare_implementation
from lifeguard.settings import LIFEGUARD_DIRECTORY, SETTINGS_MANAGER
from lifeguard.controllers import load_custom_controllers
from lifeguard.validations import load_validations

NORMAL = "NORMAL"
WARNING = "WARNING"
PROBLEM = "PROBLEM"

ACTION_STATUSES = [NORMAL, WARNING, PROBLEM]


def change_status(old, new):
    """
    Change status by severity
    """
    if ACTION_STATUSES.index(new) > ACTION_STATUSES.index(old):
        return new
    return old


def recover_settings():
    """
    Get settings.py from root of project
    """

    logger.info("lifeguard directory %s", LIFEGUARD_DIRECTORY)
    sys.path.append(LIFEGUARD_DIRECTORY)

    logger.info("path: %s", sys.path)

    lifeguard_settings = "lifeguard_settings"
    logger.info("loading %s", lifeguard_settings)
    if importlib.util.find_spec(lifeguard_settings) is None:
        logger.error("lifeguard_settings.py not found!!!")
        sys.exit(-1)
    return __import__(lifeguard_settings)


def setup(lifeguard_context):
    """
    Setup lifeguard context
    """
    # init plugins
    lifeguard_settings = recover_settings()

    for plugin in lifeguard_settings.PLUGINS:
        SETTINGS_MANAGER.settings.update(plugin.settings.SETTINGS_MANAGER.settings)
        if not lifeguard_context.only_settings:
            plugin.init(lifeguard_context)

    load_custom_controllers()
    load_validations()

    lifeguard_settings.setup(lifeguard_context)



import unittest

from unittest.mock import patch

from lifeguard import NORMAL
from lifeguard.validations import ValidationResponse, load_validations, VALIDATIONS


class TestValidationResponse(unittest.TestCase):
    def test_validation_response_object(self):
        details = {}
        settings = {}

        response = ValidationResponse("name", NORMAL, details, settings)

        self.assertEqual(response.validation_name, "name")
        self.assertEqual(response.status, "NORMAL")
        self.assertEqual(response.details, details)
        self.assertEqual(response.settings, settings)

    def test_validation_response_to_str(self):
        response = ValidationResponse("name", NORMAL, {}, {})
        self.assertEqual(
            str(response),
            "{'validation_name': 'name', 'status': 'NORMAL', 'details': {}, 'settings': {}}",
        )


class TestValidations(unittest.TestCase):
    @patch("lifeguard.validations.LIFEGUARD_DIRECTORY", "tests/fixtures")
    @patch("lifeguard.validations.logger")
    def test_load_validations(self, mock_logger):
        load_validations()
        mock_logger.info.assert_any_call("loading validation %s", "simple_validation")
        self.assertTrue("simple_validation" in VALIDATIONS)

    @patch("lifeguard.validations.LIFEGUARD_DIRECTORY", "tests/fixtures")
    @patch("lifeguard.validations.logger")
    def test_execute_validation(self, _mock_logger):
        load_validations()
        response = VALIDATIONS["simple_validation"]["ref"]()
        self.assertEqual(response.status, NORMAL)

    @patch("lifeguard.validations.LIFEGUARD_DIRECTORY", "tests/fixtures")
    @patch("lifeguard.validations.logger")
    def test_execute_validation_with_action(self, mock_logger):
        load_validations()
        VALIDATIONS["simple_with_action_validation"]["ref"]()
        mock_logger.info.assert_called_with(
            "executing action %s with result %s...",
            "simple_action",
            "{'validation_name': 'simple_with_action_validation', 'status': 'NORMAL', 'details': {}, 'settings': None}",
        )

    @patch("lifeguard.validations.LIFEGUARD_DIRECTORY", "tests/fixtures")
    @patch("lifeguard.validations.logger")
    @patch("lifeguard.validations.traceback")
    def test_execute_validation_with_invalid_action(self, mock_traceback, mock_logger):
        mock_traceback.format_exc.return_value = "traceback"
        load_validations()
        VALIDATIONS["simple_with_invalid_action_validation"]["ref"]()
        mock_logger.warning.assert_called_with(
            "validation error %s: %s",
            "simple_with_invalid_action_validation",
            "invalid_action() takes 0 positional arguments but 2 were given",
            extra={"traceback": "traceback"},
        )



"""
Actions executed after validations thats save data into database
"""
from lifeguard.repositories import ValidationRepository


def save_result_into_database(validation_response, _settings):
    """
    Save last validation result into database

    :param validation_response: a validation response
    :param _settings: not used in this action
    """

    repository = ValidationRepository()
    repository.save_validation_result(validation_response)



from setuptools import find_packages, setup

setup(
    name="lifeguard",
    version="0.0.21",
    url="https://github.com/LifeguardSystem/lifeguard",
    author="Diego Rubin",
    author_email="contact@diegorubin.dev",
    license="GPL2",
    scripts=["bin/lifeguard"],
    include_package_data=True,
    description="Application to monitor your systems and give you the security to sleep peacefully at night",
    install_requires=["flask", "schedule", "gunicorn", "requests", "tabulate"],
    classifiers=["Development Status :: 3 - Alpha"],
    packages=find_packages(),
)



"""
Base of action used to notification
"""
import json

from lifeguard import NORMAL, PROBLEM
from lifeguard.logger import lifeguard_logger as logger
from lifeguard.notifications import NOTIFICATION_METHODS, NotificationStatus
from lifeguard.repositories import NotificationRepository


def notify_in_thread(validation_response, settings):
    """
    Create a new thread to notify problem status.
    When status returns to NORMAL thread should be closed.

    :param validation_response: a validation response
    :param settings: validation settings
    """
    repository = NotificationRepository()
    last_notification_status = repository.fetch_last_notification_for_a_validation(
        validation_response.validation_name
    )

    if not last_notification_status and validation_response.status != PROBLEM:
        return

    content = json.dumps(validation_response.details)

    if not last_notification_status:
        thread_ids = {}
        for notification_method in NOTIFICATION_METHODS:
            thread_ids[notification_method.name] = notification_method.init_thread(
                content, settings
            )

        last_notification_status = NotificationStatus(
            validation_response.validation_name, thread_ids
        )
    else:
        for notification_method in NOTIFICATION_METHODS:
            if notification_method.name in last_notification_status.thread_ids:
                __send_notification_in_thread(
                    last_notification_status,
                    notification_method,
                    content,
                    settings,
                    validation_response,
                )
    repository.save_last_notification_for_a_validation(last_notification_status)


def __send_notification_in_thread(
    last_notification_status,
    notification_method,
    content,
    settings,
    validation_response,
):
    thread_id = last_notification_status.thread_ids[notification_method.name]
    if validation_response.status == NORMAL:
        notification_method.close_thread(thread_id, content, settings)
        last_notification_status.close()
    else:
        notification_method.update_thread(thread_id, content, settings)
        last_notification_status.update()


def notify_in_single_message(validation_response, settings):
    """
    This action sends single message on a new thread.

    To send notification validation response should have:
    {
        "notification": {"notify": True}
    }

    :param validation_response: a validation response
    :param settings: validation settings
    """

    response_settings = validation_response.settings or {}
    notification_settings = response_settings.get("notification", {})
    should_notify = notification_settings.get("notify", False)

    if should_notify:
        content = json.dumps(validation_response.details)
        for notification_method in NOTIFICATION_METHODS:
            notification_method.send_single_message(content, settings)



import json
import traceback
from datetime import datetime

from flask import Flask, make_response

from lifeguard.controllers import custom_controllers
from lifeguard.logger import lifeguard_logger as logger
from lifeguard.repositories import ValidationRepository
from lifeguard.validations import VALIDATIONS, ValidationResponseEncoder

APP = Flask(__name__)


def make_json_response(content):
    response = make_response(content)
    response.headers["Content-Type"] = "application/json"
    return response


@APP.route("/lifeguard/validations/<validation>", methods=["GET"])
def get_validation(validation):
    repository = ValidationRepository()
    result = repository.fetch_last_validation_result(validation)
    return make_json_response(ValidationResponseEncoder().encode(result))


@APP.route("/lifeguard/validations/<validation>/execute", methods=["POST"])
def execute_validation(validation):
    try:
        result = VALIDATIONS[validation]["ref"]()
        result.last_execution = datetime.now()
        return make_json_response(ValidationResponseEncoder().encode(result))
    except Exception:
        logger.error(
            "error on execute validation %s",
            validation,
            extra={"traceback": traceback.format_exc()},
        )
        return json.dumps({"error": traceback.format_exc()})


APP.register_blueprint(custom_controllers)
